#!/usr/local/bin/ruby
# bdrec00.rb : class for blood donation record

# Copyright (C) 1999, Masahiro SATO (m@sa.to)

require 'date2'

class BloodDonationRecord

  ##
  ## constants
  ##

  # T2 : 200ml, T4 : 400ml, PLS : Plasma, PLT : Platelet
  # DONOT : don't donate (default), WONT : will not donate
  T2, T4, PLS, PLT, PLSPLT, DONOT, WONT = 0, 1, 2, 3, 4, 5, 6

  # sex
  FEMALE, MALE = 0, 1

  # period (by week)
  T2TO24P = [4, 4] # T2 -> T2, T4, PLT, PLS
  T4TO24 = [16, 12] # T4 -> T2, T4
  T4TOP = [8, 8] # T4 -> P
  PTOP = [2, 2] # P -> P (excluding the case below)
  PLTTOPLT = [1, 1] # PLT -> PLT, use with care

  # maximum
  T_MAX = [800, 1200] # ml
  P_MAX = [24, 24] # times by plasma + 2 * platelet
#  T_MAX = { FEMALE => 800, MALE => 1200 } # ml
#  P_MAX = { FEMALE => 24, MALE => 24 } # times by plasma + 2 * platelet

  # point
  POINT = { T2 => 200, T4 => 400, PLS => 1, PLT => 2, PLSPLT => 2 }

  ##
  ## initialize
  ##
  def initialize(sex = FEMALE, calc = TRUE, allow = FALSE)
    @bd_rec = {} # hash for blood donation record
                 # key : date, value : type of donation
    @cannot_t2 = [] # array for date when cannot donate 200ml total blood
                    # value : date
    @cannot_t4 = [] # array for date when cannot donate 400ml total blood
                    # value : date
    @cannot_plt = [] # array for date when cannot donate platelet
                     # value : date
    @cannot_pls = [] # array for date when cannot donate plasma
                     # value : date
    @sex = sex # FEMALE or MALE
    @calc = calc # flag for exec calc_cannot
    @allow = allow # flag whether add the date when cannot donate or not
    @changed = FALSE # flag whether @bd_rec changed after calc_cannot
  end

  ##
  ## methods
  ##

  # dtype : return the kind of donation of the date
  #         and whether can donate or not
  def dona_type(date)
    return @bd_rec[date],
      !@cannot_t2.include?(date), !@cannot_t4.include?(date),
      !@cannot_plt.include?(date), !@cannot_pls.include?(date)
  end

  # dona_list : return donation list
  #   key : date, value : type
  def dona_list
    return @bd_rec
  end

  # cannot_list : return cannot list
  #   key : date
  #   value : array of cannot_t, cannot_plt and cannot_pls
  def cannot_list
    ha = {}
    ar = @cannot_t2 + @cannot_t4 + @cannot_plt + @cannot_pls
    ar.uniq.sort.each {|d|
      ha[d] = [
	!@cannot_t2.include?(d),
	!@cannot_t4.include?(d),
	!@cannot_plt.include?(d),
	!@cannot_pls.include?(d)
      ]
    }
    return ha
  end

  # add : add the record of donation
  def add(date, dtype)
    if @allow
      @bd_rec[date] = dtype
      @changed = TRUE
    else
      case dtype
      when T2
	unless @cannot_t2.include?(date)
	  @bd_rec[date] = dtype
	  @changed = TRUE
	end
      when T4
	unless @cannot_t4.include?(date)
	  @bd_rec[date] = dtype
	  @changed = TRUE
	end
      when PLT
	unless @cannot_plt.include?(date)
	  @bd_rec[date] = dtype
	  @changed = TRUE
	end
      when PLS
	unless @cannot_pls.include?(date)
	  @bd_rec[date] = dtype
	  @changed = TRUE
	end
      when PLSPLT
	unless @cannot_pls.include?(date) or @cannot_plt.include?(date)
	  @bd_rec[date] = dtype
	  @changed = TRUE
	end
      when DONOT
	delete(date)
      when WONT
	@bd_rec[date] = dtype
#	@changed = TRUE
      end
    end

    calc_cannot

    return dona_type(date)
  end

  # delete : delete the record of donation
  def delete(date)
    v = @bd_rec.delete(date)
    @changed = TRUE
    calc_cannot
    return v
  end

  # calc_cannot : newly create @cannot_donate hash
  def calc_cannot
    if @calc
      # tmp array
      ar_t = []
      ar_p = []
      ar_plt = []
      @bd_rec.each_pair {|key, value|
	case value
	when T2, T4
	  ar_t.push(key)
	when PLT
	  ar_p.push(key)
	  ar_plt.push(key)
	when PLS, PLSPLT
	  ar_p.push(key)
	end
      }
      ar_t.sort!
      ar_p.sort!
      ar_plt.sort!

      # clear
      @cannot_t2.clear
      @cannot_t4.clear
      @cannot_plt.clear
      @cannot_pls.clear

      # check period
      checked_t = FALSE
      checked_p = FALSE
      @bd_rec.keys.sort{|xx,yy| yy<=>xx}.each {|current|
	type = @bd_rec[current]
	if type == PLS or type == PLT or type == PLSPLT
	  next if checked_p
	  checked_p = TRUE
	elsif type == T2 or type == T4
	  next if checked_t
	  checked_t = TRUE
	end

	# to next donation
	case type
	when T2
	  w = T2TO24P[@sex]
	when T4
	  w = T4TOP[@sex]
	  t = T4TO24[@sex]
	when PLS, PLT, PLSPLT
	  w = PTOP[@sex]
	end
	if type == T2 or type == T4 or type == PLS or type == PLSPLT
	  for i in 1..(w * 7 - 1)
	    @cannot_t2.push(current + i)
	    @cannot_t4.push(current + i)
	    @cannot_plt.push(current + i)
	    @cannot_pls.push(current + i)
	  end
	elsif type == PLT
	  for i in 1..(w * 7 - 1)
	    @cannot_t2.push(current + i)
	    @cannot_t4.push(current + i)
	    @cannot_pls.push(current + i)
	  end
	  for i in 1..(PLTTOPLT[@sex] * 7 - 1)
	    @cannot_plt.push(current + i)
	  end
	end
	if type == T4
	  for i in (w * 7)..(t * 7 - 1)
	    @cannot_t2.push(current + i)
	    @cannot_t4.push(current + i)
	  end
	end

	# last 4 weeks check for PLT
	if type == PLT and
	    ar_plt.include?(current - 7) and
	    ar_plt.include?(current - 14) and
	    ar_plt.include?(current - 21)
	  for i in 1..(7 * 4 - 1)
	    @cannot_plt.push(current + i)
	  end
	end

	# last 1 year check
	if type == T2 or type == T4
	  s = find_next(ar_t, last_year(current), 0, ar_t.length - 1)
	  e = find_next(ar_t, current, s, ar_t.length - 1)
	  total = 0
	  for i in s..e
	    total += POINT[@bd_rec[ar_t[i]]]
	    if total + POINT[T4] > T_MAX[@sex]
	      j = 1
	      while ar_t[i] + j < next_year(ar_t[s])
		@cannot_t4.push(ar_t[i] + j)
		j += 1
	      end
	    end
	    if total + POINT[T2] > T_MAX[@sex]
	      j = 1
	      while ar_t[i] + j < next_year(ar_t[s])
		@cannot_t2.push(ar_t[i] + j)
		j += 1
	      end
	    end
	  end
	elsif type == PLS or type == PLT or type == PLSPLT
	  s = find_next(ar_p, last_year(current), 0, ar_p.length - 1)
	  e = find_next(ar_p, current, s, ar_p.length - 1)
	  total = 0
	  for i in s..e
	    total += POINT[@bd_rec[ar_p[i]]]
	    if total + POINT[PLT] > P_MAX[@sex]
	      j = 1
	      while ar_p[i] + j < next_year(ar_p[s])
		@cannot_plt.push(ar_p[i] + j)
		j += 1
	      end
	    end
	    if total + POINT[PLS] > P_MAX[@sex]
	      j = 1
	      while ar_p[i] + j < next_year(ar_p[s])
		@cannot_pls.push(ar_p[i] + j)
		j += 1
	      end
	    end
	  end
	end
      }

      # uniq
      [@cannot_t2, @cannot_t4, @cannot_plt, @cannot_pls].each {|c|
	c.uniq!
      }

      # delete old data
      newest = @bd_rec.keys.sort{|xx,yy| yy<=>xx}[0]
      [@cannot_t2, @cannot_t4, @cannot_plt, @cannot_pls].each {|c|
	c.delete_if {|d|
	  d <= newest
	}
      }

      # check period
      @changed = FALSE
    end
  end

  # find_next : find the next day of date in ar
  def find_next(ar, date, s, e)
    return nil if ar[e] < date
    return s if ar[s] >= date
    return e if e - s <= 1

    h = Integer((s + e) / 2)
    if ar[h] >= date
      find_next(ar, date, s, h)
    else
      find_next(ar, date, h, e)
    end
  end

  private :find_next

  # last_year : find the day of one year ago
  # assuming Date in date2.rb
  def last_year(d)
    if d.mon == 2 and d.mday == 29 # leap year
      return Date.new3(d.year - 1, 3, 1)
    else
      return Date.new3(d.year - 1, d.mon, d.mday) + 1 # ?
    end
  end

  private :last_year

  # next_year : find the day of one year later
  # assuming Date in date2.rb
  def next_year(d)
    if d.mon == 2 and d.mday == 29 # leap year
#      return Date.new3(d.year + 1, 2, 28)
      return Date.new3(d.year + 1, 3, 1)
    else
      return Date.new3(d.year + 1, d.mon, d.mday)
    end
  end

  private :next_year

  # calc_exec : let @calc TRUE
  def calc_exec
    @calc = TRUE
    return @calc # return value
  end

  # calc_stop : let @calc FALSE
  def calc_stop
    @calc = FALSE
    return @calc # return value
  end

  # calc? : return @calc
  def calc?
    return @calc # return value
  end

  # change_sex : change @sex
  def change_sex(sex = FEMALE)
    if sex == FEMALE or sex == MALE
      @sex = sex
      @changed = TRUE
      calc_cannot
      return sex # return value
    else
      return nil # return value
    end
  end

  # sex : return @sex
  def sex
    return @sex
  end

  # allow_add : let @allow TRUE
  def allow_add
    @allow = TRUE
    return @allow
  end

  # disallow_add : let @allow FALSE
  def disallow_add
    @allow = FALSE
    return @allow
  end

  # allowed? : return @allow
  def allowed?
    return @allow
  end

  # changed? : return @changed
  def changed?
    return @changed
  end
end
